local addon, shared = ...

local SILENT=		99
local RELEASE=		90
local TESTING=		20
local WDOGYIELD=	15
local FUNCTIONRUN=	10
local STATUSUPDT=	8
local INFORMATION=	5
local DETAIL=		1

---------------------
-- Variables
---------------------

_JABOB = {}
local JB = _JABOB

local _ui = { bagslots = {}, sicontrols = {}, sbcontrols = {} }
local _watchdogyield = {}
local _invbagsvisible = false

local MAXBAGSLOTS = 24
local MAXPLRBAGS = 5
local MAXBNKBAGS = 8

local _SLOTSINUSE = {}

local _BAG_SI_ROWS = 0
local _BAG_SB_ROWS = 0

local _si_scanned = false

local coroDead = false
local UINEEDED = true
local SVLOADED = false

_JABOB.SIU = _SLOTSINUSE
_JABOB.UI = _ui

---------------------
-- Functions
---------------------
local default_settings = {
	x=500, y = 500, nc = 20, si_iconsz = 32,
	bnkx=500, bnky=200, bnk_nc = 20, sb_iconsz = 32,
	version = "1.00.0001", debuglevel = SILENT
	}

local debuglevel = default_settings.debuglevel
	
local function Debug(s,l)
	if (l or INFORMATION) >= debuglevel then
		print(s)
	end
end

local yieldcnt = 0
local function WATCHDOGWAIT(x)
	local rmx = Inspect.System.Watchdog()
	while rmx < 0.03 do
		yieldcnt = yieldcnt+1
		Debug("#"..yieldcnt..":rmx: "..rmx.." ["..x.."]",WDOGYIELD)
		_watchdogyield[yieldcnt] = { tx = os.date("%Y%m%d-%H:%M:%S"), m = x, tr = rmx }
		coroutine.yield()
		rmx = Inspect.System.Watchdog()
	end
end

local function MergeTable(o,n)
	for k,v in pairs(n) do
		if type(v) == "table" then
			if o[k] == nil then
				o[k] = {}			
			end
	 	 	if type(o[k]) == 'table' then
	 			MergeTable(o[k], n[k])
	 	 	end
		else
			if o[k] == nil then
				o[k] = v
			end
		end
	end
end

function JB.ResizeBag(si)
	local sz, trl, trr, trt, trb, stype
	
	if si == "inv" then
		sz = JABOB_Settings.si_iconsz
		trl, trt, trr, trb = _ui.playerinv:GetTrimDimensions()
		stype = "si"
	elseif si == "bnk" then
		sz = JABOB_Settings.sb_iconsz
		trl, trt, trr, trb = _ui.playerbnk:GetTrimDimensions()
		stype = "sb"
	end
	
	for	k,v in pairs(_ui.bagslots) do
		if string.sub(k,1,2) == stype then
			v.frame:SetHeight(sz+4)
			v.frame:SetWidth(sz+4)
			v.icon:SetHeight(sz)
			v.icon:SetWidth(sz)
		end
	end

	if si == "inv" then
		_ui.playerinv:SetWidth((JABOB_Settings.nc * (JABOB_Settings.si_iconsz+4)) + ((JABOB_Settings.nc-1)*2) + trl+trr+4)
		_ui.playerinv:SetHeight((_BAG_SI_ROWS * (JABOB_Settings.si_iconsz+4)) + ((_BAG_SI_ROWS-1)*2) +trt+trb+4)
	else
		_ui.playerbnk:SetWidth((JABOB_Settings.bnk_nc * (JABOB_Settings.sb_iconsz+4)) + ((JABOB_Settings.bnk_nc-1)*2) + trl+trr+4)
		_ui.playerbnk:SetHeight((_BAG_SB_ROWS * (JABOB_Settings.sb_iconsz+4)) + ((_BAG_SB_ROWS-1)*2) +trt+trb+4)
	end
end

function JB.ArrangeBagSlots(sid)
	local row=0
	local cnt=0
	local ele={}
	local trl, trr, trt, trb
	local bagcnt, sitype, numcol
	
	if sid == "inv" then
		bagcnt = MAXPLRBAGS
		trl, trt, trr, trb = _ui.playerinv:GetTrimDimensions()
		sitype = "si"
		numcol = JABOB_Settings.nc
	else
		bagcnt = MAXBNKBAGS
		trl, trt, trr, trb = _ui.playerbnk:GetTrimDimensions()
		sitype = "sb"
		numcol = JABOB_Settings.bnk_nc
	end
	
	for x=1,bagcnt do
		local si = string.format("%sbg.00%d", sitype, x)
		local d = Inspect.Item.Detail(si)
		if d ~= nil then
			for y=1, d.slots do
				cnt=cnt+1
				si = string.format("%s%02d.%03d", sitype, x, y)
				_ui.bagslots[si].frame:SetVisible(true)
				if cnt==1 then
					row = row+1
					ele[row] = {}
					if row==1 then
						if sid == "inv" then
							_ui.bagslots[si].frame:SetPoint("TOPLEFT", _ui.playerinv, "TOPLEFT", trl+2,trt+2)
						else 
							_ui.bagslots[si].frame:SetPoint("TOPLEFT", _ui.playerbnk, "TOPLEFT", trl+2,trt+2)
						end
					else
						local psi = ele[row-1][1]
						_ui.bagslots[si].frame:SetPoint("TOPLEFT", _ui.bagslots[psi].frame, "BOTTOMLEFT", 0,2)
					end
				else
					local psi = ele[row][cnt-1]
					_ui.bagslots[si].frame:SetPoint("TOPLEFT", _ui.bagslots[psi].frame, "TOPRIGHT", 2,0)
				end
				ele[row][cnt] = si
				if cnt==numcol then
					cnt=0
				end
			end
		else
			for y=1, MAXBAGSLOTS do
				si = string.format("%s%02d.%03d", sitype, x, y)
				if _ui.bagslots[si] ~= nil then
					_ui.bagslots[si].frame:SetVisible(false)
				end
			end
		end
	end
	
	if sid == "inv" then
		_BAG_SI_ROWS = row
		_ui.playerinv:SetWidth((JABOB_Settings.nc * (JABOB_Settings.si_iconsz+4)) + ((JABOB_Settings.nc-1)*2) + trl+trr+4)
		_ui.playerinv:SetHeight((row * (JABOB_Settings.si_iconsz+4)) + ((row-1)*2) +trt+trb+4)
	else
		_BAG_SB_ROWS = row
		_ui.playerbnk:SetWidth((JABOB_Settings.bnk_nc * (JABOB_Settings.sb_iconsz+4)) + ((JABOB_Settings.bnk_nc-1)*2) + trl+trr+4)
		_ui.playerbnk:SetHeight((_BAG_SB_ROWS * (JABOB_Settings.sb_iconsz+4)) + ((_BAG_SB_ROWS-1)*2) +trt+trb+4)
	end
end

function JB.Event_Addon_SavedVariables_Load_End()
	if JABOB_Settings == nil then JABOB_Settings = {} end
	MergeTable(JABOB_Settings, default_settings)
	_ui.playerinv:SetPoint("TOPLEFT", UIParent, "TOPLEFT", JABOB_Settings.x, JABOB_Settings.y)
	_ui.playerbnk:SetPoint("TOPLEFT", UIParent, "TOPLEFT", JABOB_Settings.bnkx, JABOB_Settings.bnky)
	debuglevel = JABOB_Settings.debuglevel
	table.insert(Event.Item.Slot, { JB.Event_Item_Slot, addon.identifier, "Event.Item.Slot" })
	table.insert(Event.Item.Update, { JB.Event_Item_Update, addon.identifier, "Event.Item.Update" })
end

function JB.BuildBagElement(si, pf)
	local f = {}
	f.frame = UI.CreateFrame("Frame", "f.frame", pf)
	f.frame:SetVisible(false)
	f.frame:SetLayer(1)
	f.frame:SetBackgroundColor(0,0,0)
	f.icon = UI.CreateFrame("Texture", "f.icon", f.frame)
	f.icon:SetVisible(false)
	f.icon:SetLayer(2)
	f.icon:SetPoint("TOPLEFT", f.frame, "TOPLEFT", 2,2)
	f.icon:SetTexture("Rift", "Data/\\UI\\texture\\global\\placeholder_icon.dds")

	function f.icon.Event:LeftDown()
		Command.Item.Standard.Left(si)
	end
	function f.icon.Event:LeftUp()
		local _luptime = Inspect.Time.Real()
		local t,h
		t,h = Inspect.Cursor()
		if h ~= nil then
			Command.Item.Move(h, si)
			Command.Cursor(nil)
		end
	end
	function f.frame.Event:LeftUp()
		local t,h
		t,h = Inspect.Cursor()
		if h ~= nil then
			Command.Item.Move(h, si)
			Command.Cursor(nil)
		end
	end
	function f.icon.Event:RightClick()
		Command.Item.Standard.Right(si)
	end
	function f.icon.Event:MouseIn()
		local d = Inspect.Item.Detail(si)
		if d ~= nil then
			Command.Tooltip(d.id)
		end
	end
	
	f.text = UI.CreateFrame("Text", "f.text", f.frame)
	f.text:SetFontSize(12)
	f.text:SetText("")
	f.text:SetLayer(3)
	f.text:SetPoint("BOTTOMRIGHT", f.frame, "BOTTOMRIGHT", -2,-2)
	f.text:SetBackgroundColor(0.25,0.25,0.25,0.5)
	f.text:SetVisible(false)
	return f
end

function JB.BuildUI()
	_ui.context = UI.CreateContext(addon.identifier)
	
	WATCHDOGWAIT("BuildUI:1")
	_ui.playerinv = UI.CreateFrame("RiftWindow", "_ui.playerinv", _ui.context)
	_ui.playerinv:SetVisible(false)
	_ui.playerinv:SetTitle("JABOB [Player Inventory]")
	_ui.playerinv:SetLayer(0)
	
	local invborder = _ui.playerinv:GetBorder()
	
	function invborder.Event:LeftDown()
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - _ui.playerinv:GetLeft()
		self.sy = mouseData.y - _ui.playerinv:GetTop()
	end

	function invborder.Event:MouseMove()
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			_ui.playerinv:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end

	function invborder.Event:LeftUp()
		if self.MouseDown then
			self.MouseDown = false
		end
		JABOB_Settings.x = _ui.playerinv:GetLeft()
		JABOB_Settings.y = _ui.playerinv:GetTop()
	end
	
	WATCHDOGWAIT("BuildUI:2")

	_ui.sicontrols.close = UI.CreateFrame("Texture", "_ui.sicontrols.close", _ui.playerinv)
	_ui.sicontrols.close:SetPoint("TOPRIGHT", _ui.playerinv, "TOPRIGHT", -12,12)
	_ui.sicontrols.close:SetTexture(addon.identifier, "textures/cancel.png")
	_ui.sicontrols.close:SetLayer(1)
	_ui.sicontrols.close:SetHeight(32)
	_ui.sicontrols.close:SetWidth(32)
	_ui.sicontrols.close:SetVisible(true)
	
	function _ui.sicontrols.close.Event:LeftClick()
		_invbagsvisible = not _invbagsvisible
		_ui.playerinv:SetVisible(false)
	end

	_ui.sicontrols.layout = UI.CreateFrame("Texture", "_ui.sicontrols.layout", _ui.playerinv)
	_ui.sicontrols.layout:SetPoint("TOPLEFT", _ui.playerinv, "TOPLEFT", 12,12)
	_ui.sicontrols.layout:SetTexture(addon.identifier, "textures/layout.png")
	_ui.sicontrols.layout:SetLayer(10)
	_ui.sicontrols.layout:SetHeight(32)
	_ui.sicontrols.layout:SetWidth(32)
	_ui.sicontrols.layout:SetVisible(true)
	
	function _ui.sicontrols.layout.Event:WheelForward()
		JABOB_Settings.nc = JABOB_Settings.nc +1
		JB.ArrangeBagSlots("inv")
	end
	
	function _ui.sicontrols.layout.Event:WheelBack()
		JABOB_Settings.nc = JABOB_Settings.nc -1
		JB.ArrangeBagSlots("inv")
	end

	_ui.sicontrols.resize = UI.CreateFrame("Texture", "_ui.sicontrols.resize", _ui.playerinv)
	_ui.sicontrols.resize:SetPoint("TOPLEFT", _ui.sicontrols.layout, "TOPRIGHT", 2,2)
	_ui.sicontrols.resize:SetTexture(addon.identifier, "textures/shape_ungroup.png")
	_ui.sicontrols.resize:SetLayer(10)
	_ui.sicontrols.resize:SetHeight(32)
	_ui.sicontrols.resize:SetWidth(32)
	_ui.sicontrols.resize:SetVisible(true)
	
	function _ui.sicontrols.resize.Event:WheelForward()
		JABOB_Settings.si_iconsz = JABOB_Settings.si_iconsz +1
		JB.ResizeBag("inv")
	end

	function _ui.sicontrols.resize.Event:WheelBack()
		JABOB_Settings.si_iconsz = JABOB_Settings.si_iconsz -1
		JB.ResizeBag("inv")
	end
	
	WATCHDOGWAIT("BuildUI:3")

	for x=1, MAXPLRBAGS do
		for y=1, MAXBAGSLOTS do
			si = string.format("si%02d.%03d", x, y)
			WATCHDOGWAIT(string.format("BuildUI:4:%s", si))
			_ui.bagslots[si] = JB.BuildBagElement(si, _ui.playerinv)
		end
	end

	WATCHDOGWAIT("BuildUI:5")

	_ui.playerbnk = UI.CreateFrame("RiftWindow", "_ui.playerbnk", _ui.context)
	_ui.playerbnk:SetVisible(false)
	_ui.playerbnk:SetTitle("JABOB [Player Bank]")
	_ui.playerbnk:SetLayer(0)
	
	local bnkborder = _ui.playerbnk:GetBorder()
	
	function bnkborder.Event:LeftDown()
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.sx = mouseData.x - _ui.playerbnk:GetLeft()
		self.sy = mouseData.y - _ui.playerbnk:GetTop()
	end

	function bnkborder.Event:MouseMove()
		if self.MouseDown then
			local nx, ny
			local mouseData = Inspect.Mouse()
			nx = mouseData.x - self.sx
			ny = mouseData.y - self.sy
			_ui.playerbnk:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx,ny)
		end
	end

	function bnkborder.Event:LeftUp()
		if self.MouseDown then
			self.MouseDown = false
		end
		JABOB_Settings.bnkx = _ui.playerbnk:GetLeft()
		JABOB_Settings.bnky = _ui.playerbnk:GetTop()
	end
	
	WATCHDOGWAIT("BuildUI:6")

	_ui.sbcontrols.close = UI.CreateFrame("Texture", "_ui.sbcontrols.close", _ui.playerbnk)
	_ui.sbcontrols.close:SetPoint("TOPRIGHT", _ui.playerbnk, "TOPRIGHT", -12,12)
	_ui.sbcontrols.close:SetTexture(addon.identifier, "textures/cancel.png")
	_ui.sbcontrols.close:SetLayer(1)
	_ui.sbcontrols.close:SetHeight(32)
	_ui.sbcontrols.close:SetWidth(32)
	_ui.sbcontrols.close:SetVisible(true)
	
	function _ui.sbcontrols.close.Event:LeftClick()
		_ui.playerbnk:SetVisible(false)
	end

	_ui.sbcontrols.layout = UI.CreateFrame("Texture", "_ui.sbcontrols.layout", _ui.playerbnk)
	_ui.sbcontrols.layout:SetPoint("TOPLEFT", _ui.playerbnk, "TOPLEFT", 12,12)
	_ui.sbcontrols.layout:SetTexture(addon.identifier, "textures/layout.png")
	_ui.sbcontrols.layout:SetLayer(10)
	_ui.sbcontrols.layout:SetHeight(32)
	_ui.sbcontrols.layout:SetWidth(32)
	_ui.sbcontrols.layout:SetVisible(true)
	
	function _ui.sbcontrols.layout.Event:WheelForward()
		JABOB_Settings.bnk_nc = JABOB_Settings.bnk_nc +1
		JB.ArrangeBagSlots("bnk")
	end
	
	function _ui.sbcontrols.layout.Event:WheelBack()
		JABOB_Settings.bnk_nc = JABOB_Settings.bnk_nc -1
		JB.ArrangeBagSlots("bnk")
	end

	_ui.sbcontrols.resize = UI.CreateFrame("Texture", "_ui.sbcontrols.resize", _ui.playerbnk)
	_ui.sbcontrols.resize:SetPoint("TOPLEFT", _ui.sbcontrols.layout, "TOPRIGHT", 2,2)
	_ui.sbcontrols.resize:SetTexture(addon.identifier, "textures/shape_ungroup.png")
	_ui.sbcontrols.resize:SetLayer(10)
	_ui.sbcontrols.resize:SetHeight(32)
	_ui.sbcontrols.resize:SetWidth(32)
	_ui.sbcontrols.resize:SetVisible(true)
	
	function _ui.sbcontrols.resize.Event:WheelForward()
		JABOB_Settings.sb_iconsz = JABOB_Settings.sb_iconsz +1
		JB.ResizeBag("bnk")
	end

	function _ui.sbcontrols.resize.Event:WheelBack()
		JABOB_Settings.sb_iconsz = JABOB_Settings.sb_iconsz -1
		JB.ResizeBag("bnk")
	end
	
	for x=1, MAXBNKBAGS do
		for y=1, MAXBAGSLOTS do
			si = string.format("sb%02d.%03d", x, y)
			WATCHDOGWAIT(string.format("BuildUI:5:%s", si))
			_ui.bagslots[si] = JB.BuildBagElement(si, _ui.playerbnk)
		end
	end
end

function JB.EmptyBagSlot(si)
	_SLOTSINUSE[si] = nil
	_ui.bagslots[si].icon:SetVisible(false)
	_ui.bagslots[si].text:SetVisible(false)
	_ui.bagslots[si].frame:SetBackgroundColor(0,0,0)
end

function JB.PopulateBagSlot(si, d)
	_SLOTSINUSE[si] = true
	_ui.bagslots[si].icon:SetTexture("Rift", d.icon)
	_ui.bagslots[si].icon:SetVisible(true)
	if d.rarity == "uncommon" then
		_ui.bagslots[si].frame:SetBackgroundColor(0.01,0.71,0.01)
	elseif d.rarity == "rare" then
		_ui.bagslots[si].frame:SetBackgroundColor(0.13,0.42,0.81)
	elseif d.rarity == "relic" then
		_ui.bagslots[si].frame:SetBackgroundColor(0.86,0.53,0.01)
	elseif d.rarity == "epic" then
		_ui.bagslots[si].frame:SetBackgroundColor(0.60,0.26,0.88)
	elseif d.rarity == "sellable" then
		_ui.bagslots[si].frame:SetBackgroundColor(0.5,0.5,0.5)
	elseif d.rarity == nil then
		_ui.bagslots[si].frame:SetBackgroundColor(1,1,1)
	end
	
	if d.stack ~= nil then
		_ui.bagslots[si].text:SetText(string.format("%d", d.stack))
		_ui.bagslots[si].text:SetVisible(true)
	else
		_ui.bagslots[si].text:SetVisible(false)
	end
end

function JB.Event_Unit_Availability_Full(t)
	if _si_scanned then return end
	local cnt = 0
	for k,v in pairs(t) do
		if v == "player" then
			for x=1,MAXPLRBAGS do
				for y=1,MAXBAGSLOTS do
					local si = string.format("si%02d.%03d", x,y)
					if _ui.bagslots[si] ~= nil then
						local d = Inspect.Item.Detail(si)
						if d ~= nil then
							JB.PopulateBagSlot(si, d)
							cnt = cnt+1
						end
					end
				end
			end
			for x=1,MAXBNKBAGS do
				for y=1,MAXBAGSLOTS do
					local si = string.format("sb%02d.%03d", x,y)
					if _ui.bagslots[si] ~= nil then
						local d = Inspect.Item.Detail(si)
						if d ~= nil then
							JB.PopulateBagSlot(si, d)
							cnt = cnt+1
						end
					end
				end
			end
			if cnt > 0 then
				JB.ArrangeBagSlots("inv")
				JB.ArrangeBagSlots("bnk")
				JB.ResizeBag("inv")
				JB.ResizeBag("bnk")
				_si_scanned = true
			end
		end
	end
end

function JB.Event_Item_Update(t)
	for k,v in pairs(t) do
		if _ui.bagslots[k] then
			Debug(string.format("EIU: %s : %s", k, tostring(v)), TESTING)
			local d = Inspect.Item.Detail(k)
			if d ~= nil then
				JB.PopulateBagSlot(k, d)
			end
		end
	end
end

function JB.Event_Item_Slot(t)
	local rebuildneededSI = false
	local rebuildneededSB = false
	for k,v in pairs(t) do
		if _ui.bagslots[k] ~= nil then
			Debug(string.format("EIS: %s : %s", k, tostring(v)), TESTING)
			if v == false then
				JB.EmptyBagSlot(k)
			elseif v ~= "nil" then
				local d = Inspect.Item.Detail(k)
				if d ~= nil then
					JB.PopulateBagSlot(k, d)
				end						
			else
				JB.EmptyBagSlot(k)
				if string.sub(k,1,2) == "si" then
					rebuildneededSI = true
				elseif string.sub(k,1,2) == "sb" then
					rebuildneededSB = true
				end
			end
		end
	end
	if rebuildneededSI then
		JB.ArrangeBagSlots("inv")
	end
	if rebuildneededSB then
		JB.ArrangeBagSlots("bnk")
	end
end

function JB.Command_Slash_Register(args)
	local r = {}
	local numargs = 0
	for token in string.gmatch(args, "[^%s]+") do
		numargs=numargs+1
		r[numargs] = token
	end

	if r[1] == "toggle" then
		_invbagsvisible = not _invbagsvisible
		if _invbagsvisible then
			if _si_scanned == false then
				JB.Event_Unit_Availability_Full({"player"})
				_si_scanned = true
			end
		end
		_ui.playerinv:SetVisible(_invbagsvisible)
		if Inspect.Interaction("bank") then
			_ui.playerbnk:SetVisible(_invbagsvisible)
		else
			_ui.playerbnk:SetVisible(false)
		end
	end	
end

function JB.UpdateHandler()
	local si
	while true do
		if UINEEDED then
			JB.BuildUI()
			UINEEDED = false
		elseif SVLOADED then
			JB.Event_Addon_SavedVariables_Load_End()
			SVLOADED = false
		end
		coroutine.yield()
	end
end

JB.coUpdateHandler = coroutine.create(JB.UpdateHandler)

function JB.Event_System_Update_Begin()
	if coroDead then return end
	local ok, errormsg = coroutine.resume(JB.coUpdateHandler)
	if not ok then
		coroDead = true
		error("Error in coroutine: "..errormsg)
	end
end
---------------------
-- Register Events
---------------------

table.insert(Event.Addon.SavedVariables.Load.End, { function(a) if a == addon.identifier then SVLOADED = true end end, addon.identifier, "Event.Addon.SavedVariables.Load.End" })
table.insert(Command.Slash.Register("jabob"), { JB.Command_Slash_Register, addon.identifier, "Command.Slash.Register" })
table.insert(Event.System.Update.Begin, { JB.Event_System_Update_Begin, addon.identifier, "Event.System.Update.Begin" })
table.insert(Event.Unit.Availability.Full, { JB.Event_Unit_Availability_Full, addon.identifier, "Event.Unit.Availability.Full" })
print(string.format("v%s loaded.", addon.toc.Version))

--JB.BuildUI()